@@ -1,12 +1,8 @@
-"""Scalene: a high-performance, high-precision CPU *and* memory profiler for Python.
+"""Scalene: a scripting-language aware profiler for Python.
 
     https://github.com/emeryberger/scalene
 
-    Scalene uses interrupt-driven sampling for CPU profiling. For memory
-    profiling, it uses a similar mechanism but with interrupts generated
-    by a "sampling memory allocator" that produces signals everytime the
-    heap grows or shrinks by a certain amount. See libscalene.cpp for
-    details (sampling logic is in include/sampleheap.hpp).
+    See the paper "scalene-paper.pdf" in this repository for details on Scalene's design.
 
     by Emery Berger
     https://emeryberger.com
@@ -24,6 +20,7 @@ import os
 import pathlib
 import pickle
 import platform
+import random
 import signal
 import stat
 import struct
@@ -59,8 +56,8 @@ from typing import (
 )
 
 from scalene.adaptive import Adaptive
-from scalene.sparkline import SparkLine
-
+from scalene.runningstats import RunningStats
+from scalene import sparkline
 
 # Logic to ignore @profile decorators.
 try:
@@ -75,8 +72,8 @@ except AttributeError:
 
 
 assert (
-    sys.version_info[0] == 3 and sys.version_info[1] >= 5
-), "Scalene requires Python version 3.5 or above."
+    sys.version_info[0] == 3 and sys.version_info[1] >= 6
+), "Scalene requires Python version 3.6 or above."
 
 
 # Scalene currently only supports Unix-like operating systems; in
@@ -133,23 +130,23 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
         default=False,
         help="output as HTML (default: text)",
     )
+    parser.add_argument(
+        "--reduced-profile",
+        dest="reduced_profile",
+        action="store_const",
+        const=True,
+        default=False,
+        help="generate a reduced profile, with non-zero lines only (default: False).",
+    )
     parser.add_argument(
         "--profile-interval",
         type=float,
         default=float("inf"),
         help="output profiles every so many seconds.",
     )
-    parser.add_argument(
-        "--wallclock",
-        dest="wallclock",
-        action="store_const",
-        const=True,
-        default=False,
-        help="use wall clock time (default: virtual time)",
-    )
     parser.add_argument(
         "--cpu-only",
-        dest="cpuonly",
+        dest="cpu_only",
         action="store_const",
         const=True,
         default=False,
@@ -163,6 +160,14 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
         default=False,
         help="profile all executed code, not just the target program (default: only the target program)",
     )
+    parser.add_argument(
+        "--use-virtual-time",
+        dest="use_virtual_time",
+        action="store_const",
+        const=True,
+        default=False,
+        help="measure only CPU time, not time spent in I/O or blocking (default: False)",
+    )
     parser.add_argument(
         "--cpu-percent-threshold",
         dest="cpu_percent_threshold",
@@ -170,6 +175,13 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
         default=1,
         help="only report profiles with at least this percent of CPU time (default: 1%%)",
     )
+    parser.add_argument(
+        "--cpu-sampling-rate",
+        dest="cpu_sampling_rate",
+        type=float,
+        default=0.01,
+        help="CPU sampling rate (default: every 0.01s)",
+    )
     parser.add_argument(
         "--malloc-threshold",
         dest="malloc_threshold",
@@ -182,6 +194,11 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
     # Parse out all Scalene arguments and jam the remaining ones into argv.
     # https://stackoverflow.com/questions/35733262/is-there-any-way-to-instruct-argparse-python-2-7-to-remove-found-arguments-fro
     args, left = parser.parse_known_args()
+    # If the user did not enter any commands (just `scalene` or `python3 -m scalene`),
+    # print the usage information and bail.
+    if len(sys.argv) == 1:
+        parser.print_help(sys.stderr)
+        sys.exit(-1)
     return args, left
 
 
@@ -190,16 +207,16 @@ arguments, left = parse_args()
 # Load shared objects unless the user specifies "--cpu-only" at the command-line.
 # (x86-64 only for now.)
 
-if not arguments.cpuonly and (
+if not arguments.cpu_only and (
     platform.machine() != "x86_64" or struct.calcsize("P") * 8 != 64
 ):
-    arguments.cpuonly = True
+    arguments.cpu_only = True
     print(
         "scalene warning: currently only 64-bit x86-64 platforms are supported for memory and copy profiling."
     )
 
 if (
-    not arguments.cpuonly
+    not arguments.cpu_only
     and platform.machine() == "x86_64"
     and struct.calcsize("P") * 8 == 64
 ):
@@ -213,6 +230,8 @@ if (
             args = sys.argv[1:]
             args = [os.path.basename(sys.executable), "-m", "scalene"] + args
             result = subprocess.run(args)
+            if result.returncode < 0:
+                print("scalene error: received signal", signal.Signals(-result.returncode).name)
             sys.exit(result.returncode)
 
     # Similar logic, but for Mac OS X.
@@ -227,6 +246,8 @@ if (
             args = sys.argv[1:]
             args = [os.path.basename(sys.executable), "-m", "scalene"] + args
             result = subprocess.run(args, close_fds=True, shell=False)
+            if result.returncode < 0:
+                print("scalene error: received signal", signal.Signals(-result.returncode).name)
             sys.exit(result.returncode)
 
 Filename = NewType("Filename", str)
@@ -238,7 +259,7 @@ class Scalene:
     """The Scalene profiler itself."""
 
     # Debugging flag, for internal use only.
-    __debug = False
+    __debug: bool = False
 
     # We use these in is_call_function to determine whether a
     # particular bytecode is a function call.  We use this to
@@ -284,6 +305,11 @@ class Scalene:
         lambda: defaultdict(float)
     )
 
+    # Running stats for the fraction of time running on the CPU.
+    __cpu_utilization: Dict[Filename, Dict[LineNumber, RunningStats]] = defaultdict(
+        lambda: defaultdict(RunningStats)
+    )
+
     # Running count of total CPU samples per file. Used to prune reporting.
     __cpu_samples: Dict[Filename, float] = defaultdict(float)
 
@@ -331,13 +357,17 @@ class Scalene:
     __max_footprint: float = 0.0
 
     # mean seconds between interrupts for CPU sampling.
-    __mean_signal_interval: float = 0.01
+    __mean_cpu_sampling_rate: float = 0.01
 
     # last num seconds between interrupts for CPU sampling.
-    __last_signal_interval: float = __mean_signal_interval
+    __last_cpu_sampling_rate: float = __mean_cpu_sampling_rate
 
     # when did we last receive a signal?
-    __last_signal_time: float = 0
+    __last_signal_time_virtual: float = 0
+    __last_signal_time_wallclock: float = 0
+
+    # do we use wallclock time (capturing system time and blocking) or not?
+    __use_wallclock_time: bool = True
 
     # memory footprint samples (time, footprint), using 'Adaptive' sampling.
     __memory_footprint_samples = Adaptive(27)
@@ -444,7 +474,7 @@ class Scalene:
     ) -> None:
         """We replace threading.Thread.join with this method which always
 periodically yields."""
-        start_time = Scalene.gettime()
+        start_time = Scalene.get_wallclock_time()
         interval = sys.getswitchinterval()
         while self.is_alive():
             Scalene.__is_thread_sleeping[threading.get_ident()] = True
@@ -452,15 +482,15 @@ periodically yields."""
             Scalene.__is_thread_sleeping[threading.get_ident()] = False
             # If a timeout was specified, check to see if it's expired.
             if timeout:
-                end_time = Scalene.gettime()
+                end_time = Scalene.get_wallclock_time()
                 if end_time - start_time >= timeout:
                     return None
         return None
 
     @staticmethod
-    def set_timer_signal(use_wallclock_time: bool = False) -> None:
+    def set_timer_signals() -> None:
         """Set up timer signals for CPU profiling."""
-        if use_wallclock_time:
+        if Scalene.__use_wallclock_time:
             Scalene.__cpu_timer_signal = signal.ITIMER_REAL
         else:
             Scalene.__cpu_timer_signal = signal.ITIMER_VIRTUAL
@@ -471,6 +501,7 @@ periodically yields."""
         elif Scalene.__cpu_timer_signal == signal.ITIMER_VIRTUAL:
             Scalene.__cpu_signal = signal.SIGVTALRM
         elif Scalene.__cpu_timer_signal == signal.ITIMER_PROF:
+            Scalene.__cpu_signal = signal.SIGPROF
             # NOT SUPPORTED
             assert False, "ITIMER_PROF is not currently supported."
 
@@ -478,6 +509,7 @@ periodically yields."""
     def enable_signals() -> None:
         """Set up the signal handlers to handle interrupts for profiling and
 start the timer interrupts."""
+        Scalene.set_timer_signals()
         # CPU
         signal.signal(Scalene.__cpu_signal, Scalene.cpu_signal_handler)
         # Set signal handlers for memory allocation and memcpy events.
@@ -491,24 +523,28 @@ start the timer interrupts."""
         Scalene.__original_memcpy_signal_handler = signal.signal(
             Scalene.__memcpy_signal, Scalene.memcpy_event_signal_handler
         )
-        # Turn on the CPU profiling timer to run every mean_signal_interval seconds.
+        # Set every signal to restart interrupted system calls.
+        signal.siginterrupt(Scalene.__cpu_signal, False)
+        signal.siginterrupt(Scalene.__malloc_signal, False)
+        signal.siginterrupt(Scalene.__free_signal, False)
+        signal.siginterrupt(Scalene.__memcpy_signal, False)
+        # Turn on the CPU profiling timer to run every mean_cpu_sampling_rate seconds.
         signal.setitimer(
             Scalene.__cpu_timer_signal,
-            Scalene.__mean_signal_interval,
-            Scalene.__mean_signal_interval,
+            Scalene.__mean_cpu_sampling_rate,
+            Scalene.__mean_cpu_sampling_rate,
         )
-        Scalene.__last_signal_time = Scalene.gettime()
+        Scalene.__last_signal_time_virtual = Scalene.get_process_time()
 
     @staticmethod
-    def gettime() -> float:
-        """High-precision timer of time spent running in or on behalf of this
-process."""
-        if Scalene.__cpu_timer_signal == signal.ITIMER_VIRTUAL:
-            # Using virtual time
-            return time.process_time()
-        else:
-            # Using wall clock time
-            return time.perf_counter()
+    def get_process_time() -> float:
+        """Time spent on the CPU."""
+        return time.process_time()
+
+    @staticmethod
+    def get_wallclock_time() -> float:
+        """Wall-clock time."""
+        return time.perf_counter()
 
     class ReplacementLock(object):
         """Replace lock with a version that periodically yields and updates sleeping status."""
@@ -520,7 +556,7 @@ process."""
             tident = threading.get_ident()
             if blocking == 0:
                 blocking = False
-            start_time = Scalene.gettime()
+            start_time = Scalene.get_wallclock_time()
             if blocking:
                 if timeout < 0:
                     interval = sys.getswitchinterval()
@@ -538,7 +574,7 @@ process."""
                     return False
                 # If a timeout was specified, check to see if it's expired.
                 if timeout != -1:
-                    end_time = Scalene.gettime()
+                    end_time = Scalene.get_wallclock_time()
                     if end_time - start_time >= timeout:
                         return False
 
@@ -570,6 +606,10 @@ process."""
             Scalene.__cpu_percent_threshold = int(arguments.cpu_percent_threshold)
         if "malloc_threshold" in arguments:
             Scalene.__malloc_threshold = int(arguments.malloc_threshold)
+        if "cpu_sampling_rate" in arguments:
+            Scalene.__mean_cpu_sampling_rate = float(arguments.cpu_sampling_rate)
+        if arguments.use_virtual_time:
+            Scalene.__use_wallclock_time = False
 
         if arguments.pid:
             # Child process.
@@ -583,12 +623,15 @@ process."""
         else:
             # Parent process.
             # Create a temporary directory to hold aliases to the Python
-            # executable, so scalene can handle multiple proceses; each
+            # executable, so scalene can handle multiple processes; each
             # one is a shell script that redirects to Scalene.
             cmdline = ""
             preface = ""
             # Pass along commands from the invoking command line.
-            if arguments.cpuonly:
+            cmdline += " --cpu-sampling-rate=" + str(arguments.cpu_sampling_rate)
+            if arguments.use_virtual_time:
+                cmdline += " --use-virtual-time"
+            if arguments.cpu_only:
                 cmdline += " --cpu-only"
             else:
                 preface = "PYTHONMALLOC=malloc "
@@ -623,6 +666,8 @@ process."""
             os.environ["PATH"] = (
                 Scalene.__python_alias_dir_name + ":" + os.environ["PATH"]
             )
+            # Force the executable (if anyone invokes it later) to point to one of our aliases.
+            sys.executable = Scalene.__all_python_names[0]
 
         # Register the exit handler to run when the program terminates or we quit.
         atexit.register(Scalene.exit_handler)
@@ -651,9 +696,10 @@ process."""
         """Handle interrupts for CPU profiling."""
         # Record how long it has been since we received a timer
         # before.  See the logic below.
-        now = Scalene.gettime()
+        now_virtual = Scalene.get_process_time()
+        now_wallclock = Scalene.get_wallclock_time()
         # If it's time to print some profiling info, do so.
-        if now >= Scalene.__next_output_time:
+        if now_virtual >= Scalene.__next_output_time:
             # Print out the profile. Set the next output time, stop
             # signals, print the profile, and then start signals
             # again.
@@ -679,13 +725,22 @@ process."""
         # account for this time by tracking the elapsed (process) time
         # and compare it to the interval, and add any computed delay
         # (as if it were sampled) to the C counter.
-        elapsed = now - Scalene.__last_signal_time
-        python_time = Scalene.__last_signal_interval
-        c_time = elapsed - python_time
+        elapsed_virtual = now_virtual - Scalene.__last_signal_time_virtual
+        elapsed_wallclock = now_wallclock - Scalene.__last_signal_time_wallclock
+        # CPU utilization is the fraction of time spent on the CPU
+        # over the total wallclock time.
+        cpu_utilization = elapsed_virtual / elapsed_wallclock
+        if cpu_utilization > 1.0:
+            # Sometimes, for some reason, virtual time exceeds
+            # wallclock time, which makes no sense...
+            cpu_utilization = 1.0
+        assert 0 <= cpu_utilization <= 1, "CPU utilization must be between 0 and 1."
+        python_time = Scalene.__last_cpu_sampling_rate
+        c_time = elapsed_virtual - python_time
         if c_time < 0:
             c_time = 0
-        # Update counters for every running thread.
 
+        # Update counters for every running thread.
         new_frames = Scalene.compute_frames_to_record(this_frame)
         # Now update counters (weighted) for every frame we are tracking.
         total_time = python_time + c_time
@@ -715,6 +770,7 @@ process."""
                     Scalene.__cpu_samples[fname] += (
                         python_time + c_time
                     ) / total_frames
+                    Scalene.__cpu_utilization[fname][lineno].push(cpu_utilization)
             else:
                 # We can't play the same game here of attributing
                 # time, because we are in a thread, and threads don't
@@ -732,11 +788,28 @@ process."""
                         # Not in a call function so we attribute the time to Python.
                         Scalene.__cpu_samples_python[fname][lineno] += normalized_time
                     Scalene.__cpu_samples[fname] += normalized_time
+                    Scalene.__cpu_utilization[fname][lineno].push(cpu_utilization)
 
         del new_frames
 
         Scalene.__total_cpu_samples += total_time
-        Scalene.__last_signal_time = Scalene.gettime()
+        # Pick a new random interval, distributed around the mean.
+        next_interval = 0.0
+        while next_interval <= 0.0:
+            # Choose a normally distributed random number around the
+            # mean for the next interval. By setting the standard
+            # deviation to a fraction of the mean, we know by
+            # properties of the normal distribution that the
+            # likelihood of iterating this loop more than once is
+            # low. For a fraction 1/f, the probability is
+            # p = 1-(math.erf(f/math.sqrt(2)))/2
+            next_interval = random.normalvariate(
+                Scalene.__mean_cpu_sampling_rate, Scalene.__mean_cpu_sampling_rate / 3.0
+            )
+        Scalene.__last_cpu_sampling_rate = next_interval
+        Scalene.__last_signal_time_wallclock = Scalene.get_wallclock_time()
+        Scalene.__last_signal_time_virtual = Scalene.get_process_time()
+        signal.setitimer(Scalene.__cpu_timer_signal, next_interval, next_interval)
 
     # Returns final frame (up to a line in a file we are profiling), the thread identifier, and the original frame.
     @staticmethod
@@ -953,6 +1026,8 @@ process."""
     @lru_cache(None)
     def should_trace(filename: str) -> bool:
         """Return true if the filename is one we should trace."""
+        if not filename:
+            return False
         if filename[0] == "<":
             # Not a real file.
             return False
@@ -974,19 +1049,19 @@ process."""
     def start() -> None:
         """Initiate profiling."""
         Scalene.enable_signals()
-        Scalene.__start_time = Scalene.gettime()
+        Scalene.__start_time = Scalene.get_wallclock_time()
 
     @staticmethod
     def stop() -> None:
         """Complete profiling."""
         Scalene.disable_signals()
-        Scalene.__elapsed_time += Scalene.gettime() - Scalene.__start_time
+        Scalene.__elapsed_time += Scalene.get_wallclock_time() - Scalene.__start_time
 
     @staticmethod
     def output_profile_line(
         fname: Filename, line_no: LineNumber, line: str, console: Console, tbl: Table,
-    ) -> None:
-        """Print exactly one line of the profile."""
+    ) -> bool:
+        """Print at most one line of the profile (true == printed one)."""
         current_max = Scalene.__max_footprint
         did_sample_memory: bool = (
             Scalene.__total_memory_free_samples + Scalene.__total_memory_malloc_samples
@@ -1043,24 +1118,39 @@ process."""
         n_python_fraction = 0 if not n_malloc_mb else n_python_malloc_mb / n_malloc_mb
         # Finally, print results.
         n_cpu_percent_c_str: str = (
-            "" if not n_cpu_percent_c else "%6.0f%%" % n_cpu_percent_c
+            "" if n_cpu_percent_c < 0.5 else "%6.0f%%" % n_cpu_percent_c
         )
         n_cpu_percent_python_str: str = (
-            "" if not n_cpu_percent_python else "%6.0f%%" % n_cpu_percent_python
+            "" if n_cpu_percent_python < 0.5 else "%5.0f%%" % n_cpu_percent_python
         )
         n_growth_mb_str: str = (
             "" if (not n_growth_mb and not n_usage_fraction) else "%5.0f" % n_growth_mb
         )
         n_usage_fraction_str: str = (
-            "" if not n_usage_fraction else "%3.0f%%" % (100 * n_usage_fraction)
+            "" if n_usage_fraction < 0.5 else "%3.0f%%" % (100 * n_usage_fraction)
         )
         n_python_fraction_str: str = (
-            "" if not n_python_fraction else "%5.0f%%" % (100 * n_python_fraction)
+            "" if n_python_fraction < 0.5 else "%5.0f%%" % (100 * n_python_fraction)
         )
         n_copy_b = Scalene.__memcpy_samples[fname][line_no]
         n_copy_mb_s = n_copy_b / (1024 * 1024 * Scalene.__elapsed_time)
         n_copy_mb_s_str: str = "" if n_copy_mb_s < 0.5 else "%6.0f" % n_copy_mb_s
 
+        n_cpu_percent = n_cpu_percent_c + n_cpu_percent_python
+        # Only report utilization where there is more than 1% CPU total usage,
+        # and the standard error of the mean is low (meaning it's an accurate estimate).
+        sys_str: str = "" if n_cpu_percent < 0.5 or Scalene.__cpu_utilization[fname][
+            line_no
+        ].size() <= 1 or Scalene.__cpu_utilization[fname][
+            line_no
+        ].sem() > 0.025 or Scalene.__cpu_utilization[
+            fname
+        ][
+            line_no
+        ].mean() > 0.99 else "%3.0f%%" % (
+            n_cpu_percent * (1.0 - (Scalene.__cpu_utilization[fname][line_no].mean()))
+        )
+
         if did_sample_memory:
             spark_str: str = ""
             # Scale the sparkline by the usage fraction.
@@ -1068,7 +1158,7 @@ process."""
             for i in range(0, len(samples.get())):
                 samples.get()[i] *= n_usage_fraction
             if samples.get():
-                _, _, spark_str = SparkLine().generate(
+                _, _, spark_str = sparkline.generate(
                     samples.get()[0 : samples.len()], 0, current_max
                 )
 
@@ -1089,16 +1179,21 @@ process."""
                 ncpcs = n_cpu_percent_c_str
                 nufs = spark_str + n_usage_fraction_str
 
-            tbl.add_row(
-                str(line_no),
-                ncpps,  # n_cpu_percent_python_str,
-                ncpcs,  # n_cpu_percent_c_str,
-                n_python_fraction_str,
-                n_growth_mb_str,
-                nufs,  # spark_str + n_usage_fraction_str,
-                n_copy_mb_s_str,
-                syntax_highlighted,
-            )
+            if not arguments.reduced_profile or ncpps + ncpcs + nufs:
+                tbl.add_row(
+                    str(line_no),
+                    ncpps,  # n_cpu_percent_python_str,
+                    ncpcs,  # n_cpu_percent_c_str,
+                    sys_str,
+                    n_python_fraction_str,
+                    n_growth_mb_str,
+                    nufs,  # spark_str + n_usage_fraction_str,
+                    n_copy_mb_s_str,
+                    syntax_highlighted,
+                )
+                return True
+            else:
+                return False
 
         else:
 
@@ -1112,12 +1207,17 @@ process."""
                 ncpps = n_cpu_percent_python_str
                 ncpcs = n_cpu_percent_c_str
 
-            tbl.add_row(
-                str(line_no),
-                ncpps,  # n_cpu_percent_python_str,
-                ncpcs,  # n_cpu_percent_c_str,
-                syntax_highlighted,
-            )
+            if not arguments.reduced_profile or ncpps + ncpcs:
+                tbl.add_row(
+                    str(line_no),
+                    ncpps,  # n_cpu_percent_python_str,
+                    ncpcs,  # n_cpu_percent_c_str,
+                    sys_str,
+                    syntax_highlighted,
+                )
+                return True
+            else:
+                return False
 
     @staticmethod
     def output_stats(pid: int) -> None:
@@ -1211,7 +1311,7 @@ process."""
             # Nothing to output.
             return False
         # Collect all instrumented filenames.
-        all_instrumented_files: List[str] = list(
+        all_instrumented_files: List[Filename] = list(
             set(
                 list(Scalene.__cpu_samples_python.keys())
                 + list(Scalene.__cpu_samples_c.keys())
@@ -1232,7 +1332,7 @@ process."""
             samples = Scalene.__memory_footprint_samples
             if len(samples.get()) > 0:
                 # Output a sparkline as a summary of memory usage over time.
-                _, _, spark_str = SparkLine().generate(
+                _, _, spark_str = sparkline.generate(
                     samples.get()[0 : samples.len()], 0, current_max
                 )
                 mem_usage_line = Text.assemble(
@@ -1245,8 +1345,11 @@ process."""
         null = open("/dev/null", "w")
         console = Console(width=132, record=True, force_terminal=True, file=null)
         # Build a list of files we will actually report on.
-        report_files = []
-        for fname in sorted(all_instrumented_files):
+        report_files: List[Filename] = []
+        # Sort in descending order of CPU cycles, and then ascending order by filename
+        for fname in sorted(
+            all_instrumented_files, key=lambda f: (-(Scalene.__cpu_samples[f]), f)
+        ):
             fname = Filename(fname)
             try:
                 percent_cpu_time = (
@@ -1275,7 +1378,7 @@ process."""
                 100 * Scalene.__cpu_samples[fname] / Scalene.__total_cpu_samples
             )
             new_title = mem_usage_line + (
-                "%s: %% of CPU time = %6.2f%% out of %6.2fs."
+                "%s: %% of time = %6.2f%% out of %6.2fs."
                 % (fname, percent_cpu_time, Scalene.__elapsed_time)
             )
             # Only display total memory usage once.
@@ -1286,22 +1389,33 @@ process."""
             )
 
             tbl.add_column("Line", justify="right", no_wrap=True)
-            tbl.add_column("CPU %\nPython", no_wrap=True)
-            tbl.add_column("CPU %\nnative", no_wrap=True)
+            tbl.add_column("Time %\nPython", no_wrap=True)
+            tbl.add_column("Time %\nnative", no_wrap=True)
+            tbl.add_column("Sys\n%", no_wrap=True)
             if did_sample_memory:
                 tbl.add_column("Mem %\nPython", no_wrap=True)
                 tbl.add_column("Net\n(MB)", no_wrap=True)
                 tbl.add_column("Memory usage\nover time / %", no_wrap=True)
                 tbl.add_column("Copy\n(MB/s)", no_wrap=True)
-                tbl.add_column("\n" + fname, width=66)
+                tbl.add_column("\n" + fname, width=60)
             else:
                 tbl.add_column("\n" + fname, width=96)
 
+            # Print out the the profile for the source, line by line.
             with open(fname, "r") as source_file:
+                # We track whether we should put in ellipsis (for reduced profiles)
+                # or not.
+                did_print = True  # did we print a profile line last time?
                 for line_no, line in enumerate(source_file, 1):
-                    Scalene.output_profile_line(
+                    old_did_print = did_print
+                    did_print = Scalene.output_profile_line(
                         fname, LineNumber(line_no), line, console, tbl
                     )
+                    if old_did_print and not did_print:
+                        # We are skipping lines, so add an ellipsis.
+                        tbl.add_row("...")
+                    old_did_print = did_print
+
             console.print(tbl)
 
         if Scalene.__html:
@@ -1343,66 +1457,73 @@ process."""
     @staticmethod
     def main() -> None:
         """Invokes the profiler from the command-line."""
-        args, left = parse_args()  # We currently do this twice, but who cares.
-        sys.argv = left
-        Scalene.set_timer_signal(args.wallclock)
-        Scalene.__output_profile_interval = args.profile_interval
-        Scalene.__next_output_time = (
-            Scalene.gettime() + Scalene.__output_profile_interval
-        )
-        Scalene.__html = args.html
-        Scalene.__output_file = args.outfile
-        Scalene.__profile_all = args.profile_all
         try:
-            with open(sys.argv[0], "rb") as prog_being_profiled:
-                # Read in the code and compile it.
-                code = compile(prog_being_profiled.read(), sys.argv[0], "exec")
-                # Push the program's path.
-                program_path = os.path.dirname(os.path.abspath(sys.argv[0]))
-                sys.path.insert(0, program_path)
-                Scalene.__program_path = program_path
-                # Grab local and global variables.
-                import __main__
-
-                the_locals = __main__.__dict__
-                the_globals = __main__.__dict__
-                # Splice in the name of the file being executed instead of the profiler.
-                the_globals["__file__"] = os.path.basename(sys.argv[0])
-                # Start the profiler.
-                fullname = os.path.join(program_path, os.path.basename(sys.argv[0]))
-                profiler = Scalene(Filename(fullname))
-                try:
-                    # We exit with this status (returning error code as appropriate).
-                    exit_status = 0
-                    profiler.start()
-                    # Run the code being profiled.
+            args, left = parse_args()  # We currently do this twice, but who cares.
+            sys.argv = left
+            Scalene.__output_profile_interval = args.profile_interval
+            Scalene.__next_output_time = (
+                Scalene.get_wallclock_time() + Scalene.__output_profile_interval
+            )
+            Scalene.__html = args.html
+            Scalene.__output_file = args.outfile
+            Scalene.__profile_all = args.profile_all
+            try:
+                with open(sys.argv[0], "rb") as prog_being_profiled:
+                    # Read in the code and compile it.
                     try:
-                        exec(code, the_globals, the_locals)
-                    except SystemExit as se:
-                        # Intercept sys.exit and propagate the error code.
-                        exit_status = se.code
-                    except BaseException:
-                        if Scalene.__debug:
-                            print(traceback.format_exc())  # for debugging only
-                        pass
-                    profiler.stop()
-                    # If we've collected any samples, dump them.
-                    if profiler.output_profiles():
-                        pass
-                    else:
-                        print(
-                            "Scalene: Program did not run for long enough to profile."
-                        )
-                    sys.exit(exit_status)
-                except Exception as ex:
-                    template = (
-                        "Scalene: An exception of type {0} occurred. Arguments:\n{1!r}"
-                    )
-                    message = template.format(type(ex).__name__, ex.args)
-                    print(message)
-                    print(traceback.format_exc())
-        except (FileNotFoundError, IOError):
-            print("Scalene: could not find input file " + sys.argv[0])
+                        code = compile(prog_being_profiled.read(), sys.argv[0], "exec")
+                    except SyntaxError:
+                        traceback.print_exc()
+                        sys.exit(-1)
+                    # Push the program's path.
+                    program_path = os.path.dirname(os.path.abspath(sys.argv[0]))
+                    sys.path.insert(0, program_path)
+                    Scalene.__program_path = program_path
+                    # Grab local and global variables.
+                    import __main__
+
+                    the_locals = __main__.__dict__
+                    the_globals = __main__.__dict__
+                    # Splice in the name of the file being executed instead of the profiler.
+                    the_globals["__file__"] = os.path.basename(sys.argv[0])
+                    # Start the profiler.
+                    fullname = os.path.join(program_path, os.path.basename(sys.argv[0]))
+                    profiler = Scalene(Filename(fullname))
+                    try:
+                        # We exit with this status (returning error code as appropriate).
+                        exit_status = 0
+                        profiler.start()
+                        # Run the code being profiled.
+                        try:
+                            exec(code, the_globals, the_locals)
+                        except SystemExit as se:
+                            # Intercept sys.exit and propagate the error code.
+                            exit_status = se.code
+                        except BaseException:
+                            if Scalene.__debug:
+                                print(traceback.format_exc())  # for debugging only
+                            pass
+                        profiler.stop()
+                        # If we've collected any samples, dump them.
+                        if profiler.output_profiles():
+                            pass
+                        else:
+                            print(
+                                "Scalene: Program did not run for long enough to profile."
+                            )
+                        sys.exit(exit_status)
+                    except Exception as ex:
+                        template = "Scalene: An exception of type {0} occurred. Arguments:\n{1!r}"
+                        message = template.format(type(ex).__name__, ex.args)
+                        print(message)
+                        print(traceback.format_exc())
+            except (FileNotFoundError, IOError):
+                print("Scalene: could not find input file " + sys.argv[0])
+                sys.exit(-1)
+        except SystemExit as se:
+            pass
+        except BaseException:
+            print("Scalene failed to initialize.\n" + traceback.format_exc())
             sys.exit(-1)
 
 
